module set_all_outerm_m

  implicit none

contains

  subroutine set_all_outerm(s, min_amp, max_radius, step, periodic, ssh, &
       corner, min_area)

    ! Extraction of eddies: find extrema and set all outermost
    ! contours in the snapshot "s". Not a function because "s" is not
    ! completely defined.

    ! Libraries:
    use jumble, only: argwhere
    use numer_rec_95, only: indexx
    use nr_util, only: twopi

    use derived_types, only: snapshot
    use get_1_outerm_m, only: get_1_outerm
    use nearby_extr_m, only: nearby_extr
    use local_extrema_m, only: local_extrema

    type(snapshot), intent(out):: s
    ! Specifically: define everything in s except
    ! s%list_vis%speed_cont, s%list_vis%max_speed, s%list_vis%radius4
    ! and s%number_eddies.

    real, intent(in):: min_amp ! minimum amplitude of ssh, between
    ! extremum and outermost contour, in m

    integer, intent(in):: max_radius(:) ! (2) maximum radius of an
    ! eddy in longitude and latitude, in number of grid points

    real, intent(in):: step(:) ! (2) longitude and latitude steps, in rad
    logical, intent(in):: periodic ! grid is periodic in longitude

    real, intent(in):: ssh(1 - merge(max_radius(1), 0, periodic):, :)
    ! (1 - max_radius(1):nlon + max_radius(1), nlat) if the grid is periodic
    ! in longitude, else (nlon, nlat). Sea-surface height, in m.

    real, intent(in):: corner(:) ! (2) longitude and latitude of the
    ! corner of the global grid, in rad

    real, intent(in):: min_area
    ! minimum area of an outermost contour, in m2
    
    ! Local:

    integer nlon, nlat

    real, allocatable:: innermost_level(:) ! (s%number_vis_extr)
    ! SSH level of the innermost contour around each extremum, in
    ! m. By construction, innermost_level < extremum for a maximum, >
    ! extremum for a minimum.

    logical, allocatable:: cyclone(:) ! (s%number_vis_extr)
    integer i, l, copy
    integer n_cycl ! number of cyclones

    integer, allocatable:: sorted_extr(:) ! (s%number_vis_extr)
    ! Sorted identifying number of extrema: first, minima sorted in
    ! order of decreasing SSH, and second, maxima sorted in order of
    ! increasing SSH.

    integer, allocatable:: selection(:)
    ! identifying numbers of a selection of extrema

    ! Window around each extremum:

    integer llc(2) ! indices in global grid of lower left corner
    integer urc(2) ! indices in global grid of upper right corner

    real corner_window(2) ! longitude and latitude of the window around each
    ! extremum, in rad

    real, allocatable:: outside_points(:, :) ! (2, :) longitude and
    ! latitude, in rad, of all the significant extrema, except the
    ! target extremum

    !--------------------------------------------------------------

    copy = merge(max_radius(1), 0, periodic)
    nlon = size(ssh, 1) - 2 * copy
    nlat = size(ssh, 2)
    allocate(s%extr_map(1 - copy:nlon + copy, nlat))
    copy = merge(1, 0, periodic)
    call local_extrema(s%extr_map(1 - copy:nlon + copy, :), s%ind_extr, &
         innermost_level, cyclone, ssh(1 - copy:nlon + copy, :), periodic, &
         my_lbound = [1 - copy, 1])
    ! Assertion: \forall k, s%ind_extr(1, k) \in \{2 - copy, \dots,
    ! nlon + copy - 1\} et s%ind_extr(2, k) \in \{2, \dots, nlat - 1\}
    
    ! Assertion: \forall k, s%ind_extr(1, k) \in \{1, \dots, nlon\} et
    ! s%ind_extr(2, k) \in \{2, \dots, nlat - 1\}

    if (periodic) then
       ! Extend in longitude:
       s%extr_map(1 - max_radius(1):- 1, :) &
            = s%extr_map(nlon + 1 - max_radius(1):nlon - 1, :)
       s%extr_map(nlon + 2:nlon + max_radius(1), :) &
            = s%extr_map(2:max_radius(1), :)
    end if

    s%number_vis_extr = size(s%ind_extr, 2)
    allocate(s%list_vis(s%number_vis_extr), sorted_extr(s%number_vis_extr))

    forall (i = 1:s%number_vis_extr)
       s%list_vis(i)%coord_extr = corner + (s%ind_extr(:, i) - 1) * step
       ! (Even when periodic, this is within the original NetCDF grid,
       ! that is the grid without duplicated longitudes.)
       
       s%list_vis(i)%ssh_extr = ssh(s%ind_extr(1, i), s%ind_extr(2, i))
       s%list_vis(i)%cyclone = cyclone(i)
       s%list_vis(i)%interpolated = .false.
       s%list_vis(i)%valid = .true.
    end forall

    ! Double sort of extrema on cyclonicity and SSH value:

    selection = argwhere(cyclone)
    n_cycl = size(selection)
    selection = selection(indexx(s%list_vis(selection)%ssh_extr))
    sorted_extr(:n_cycl) = selection(n_cycl:1:- 1) ! descending order of ssh

    selection = argwhere(.not. cyclone)
    selection = selection(indexx(s%list_vis(selection)%ssh_extr))
    sorted_extr(n_cycl + 1:) = selection

    do l = 1, s%number_vis_extr
       i = sorted_extr(l)

       ! Define the geographical window around each eddy extremum:

       llc = s%ind_extr(:, i) - max_radius
       urc = s%ind_extr(:, i) + max_radius

       llc(2) = max(llc(2), 1)
       urc(2) = min(urc(2), nlat)

       if (.not. periodic) then
          llc(1) = max(llc(1), 1)
          urc(1) = min(urc(1), nlon)
       end if

       corner_window = corner + (llc - 1) * step

       ! No need to consider contours with amplitudes < min_amp:
       if (abs(s%list_vis(i)%ssh_extr - innermost_level(i)) < min_amp) &
            innermost_level(i) = s%list_vis(i)%ssh_extr &
            + merge(min_amp, - min_amp, s%list_vis(i)%cyclone)
       
       outside_points = nearby_extr(s%extr_map(llc(1):urc(1), llc(2):urc(2)), &
            s%list_vis, i)
       if (periodic) outside_points(1, :) = outside_points(1, :) &
            + ceiling((corner_window(1) - outside_points(1, :)) / twopi) * twopi
       ! (Shift the longitude of each point to a value close to the
       ! longitude of the target extremum.)
       s%list_vis(i)%out_cont = get_1_outerm(s%list_vis(i)%cyclone, &
            s%list_vis(i)%coord_extr, innermost_level(i), outside_points, &
            ssh(llc(1):urc(1), llc(2):urc(2)), corner_window, step)
       s%list_vis(i)%valid = s%list_vis(i)%out_cont%area >= min_area
    end do

  end subroutine set_all_outerm

end module set_all_outerm_m
