module successive_overlap_m

  implicit none

contains

  subroutine successive_overlap(flow, unit_edgelist, nlon, nlat, periodic, &
       dist_lim, j, k)

    ! Finds edges between flow(j - 1) and flow(j), corresponding to
    ! dates k - 1 and k. Writes these edges to unit_edgelist. Updates
    ! flow(j - 1)%list_vis%delta_out and flow(j)%list_vis%delta_in.

    ! Libraries:
    use contour_531, only: polyline
    use gpc_f, only: gpc_polygon_clip_f, GPC_INT, polygon
    use nr_util, only: twopi

    use candidate_overlap_m, only: candidate_overlap
    use derived_types, only: snapshot
    use spher_polygon_area_m, only: spher_polygon_area
    use spher_polyline_area_m, only: spher_polyline_area
    use weight_m, only: weight

    type(snapshot), intent(inout):: flow(:) ! (max_delta + 1)
    integer, intent(in):: unit_edgelist ! logical unit for edgelist file
    integer, intent(in):: nlon, nlat
    logical, intent(in):: periodic ! grid is periodic in longitude

    integer, intent(in):: dist_lim
    ! We look for an overlapping eddy at dist_lim (in grid points) of
    ! the first extremum.

    integer, intent(in):: j, k

    ! Local:

    integer i1, i2, l, n_select, m
    type(polyline) polyline_1, polyline_2
    type(polygon) res_pol

    integer, allocatable:: selection(:)
    ! identifying numbers of a selection of extrema

    ! Window around each extremum:
    integer llc(2) ! indices in global grid of lower left corner
    integer urc(2) ! indices in global grid of upper right corner

    !-----------------------------------------------------------------------

    do i1 = 1, flow(j - 1)%number_vis_extr
       if (flow(j - 1)%list_vis(i1)%valid) then
          ! Define the geographical window around each eddy extremum:

          llc = flow(j - 1)%ind_extr(:, i1) - dist_lim
          urc = flow(j - 1)%ind_extr(:, i1) + dist_lim

          llc(2) = max(llc(2), 1)
          urc(2) = min(urc(2), nlat)

          if (.not. periodic) then
             llc(1) = max(llc(1), 1)
             urc(1) = min(urc(1), nlon)
          end if

          selection = candidate_overlap(flow(j)%extr_map(llc(1):urc(1), &
               llc(2):urc(2)), flow(j)%list_vis, &
               flow(j - 1)%list_vis(i1)%cyclone)
          n_select = size(selection)
          if (n_select /= 0) polyline_1 &
               = merge(flow(j - 1)%list_vis(i1)%speed_cont%polyline, &
               flow(j - 1)%list_vis(i1)%out_cont%polyline, &
               flow(j - 1)%list_vis(i1)%speed_cont%n_points /= 0)

          DO l = 1, n_select
             i2 = selection(l)
             polyline_2 = merge(flow(j)%list_vis(i2)%speed_cont%polyline, &
                  flow(j)%list_vis(i2)%out_cont%polyline, &
                  flow(j)%list_vis(i2)%speed_cont%n_points /= 0)

             ! Shift the longitudes of polyline_2 to values close to the
             ! longitude of extremum i1:
             m = floor((flow(j - 1)%list_vis(i1)%coord_extr(1) &
                  - flow(j)%list_vis(i2)%coord_extr(1)) / twopi + 0.5)
             polyline_2%points(1, :) = polyline_2%points(1, :) + m * twopi

             call gpc_polygon_clip_f(GPC_INT, polygon(nparts = 1, &
                  part = [polyline_1], hole = [.false.]), &
                  polygon(nparts = 1, part = [polyline_2], hole = [.false.]), &
                  res_pol)

             if (res_pol%nparts /= 0) then
                ! polyline_1 and polyline_2 overlap
                if (spher_polygon_area(res_pol) >= 0.25 &
                     * min(spher_polyline_area(polyline_1), &
                     spher_polyline_area(polyline_2))) then
                   write(unit_edgelist, fmt = *) k - 1, i1, k, i2, &
                        weight(flow(j - 1)%list_vis(i1), flow(j)%list_vis(i2))
                   flow(j - 1)%list_vis(i1)%delta_out = 1
                   flow(j)%list_vis(i2)%delta_in = 1
                end if
             end if
          end DO
       end if
    end do

  end subroutine successive_overlap
  
end module successive_overlap_m
